@@ -19,14 +19,15 @@ from google.protobuf.field_mask_pb2 import FieldMask
 
 import nbox.utils as U
 from nbox.auth import secret
-from nbox.utils import logger
+from nbox.utils import logger, SimplerTimes
 from nbox.version import __version__
 from nbox.hyperloop.dag_pb2 import DAG
 from nbox.init import nbox_ws_v1, nbox_grpc_stub
 from nbox.hyperloop.job_pb2 import NBXAuthInfo, Job as JobProto, Resource
-from nbox.messages import rpc, write_string_to_file, get_current_timestamp
+from nbox.messages import rpc, write_binary_to_file
 from nbox.jobs import Schedule, _get_job_data, _get_deployment_data, JobInfo, Serve, Job
 from nbox.hyperloop.nbox_ws_pb2 import UploadCodeRequest, CreateJobRequest, UpdateJobRequest
+from nbox.nbxlib.operator_spec import OperatorType as OT
 
 
 #######################################################################################################################
@@ -57,14 +58,21 @@ def deploy_serving(
   if wait_for_deployment:
     logger.warning("Wait for deployment is coming in the following release!")
 
-  serving_id, serving_name = _get_deployment_data(deployment_id_or_name, workspace_id)
+  serving_id, serving_name = _get_deployment_data(deployment_id_or_name, workspace_id = workspace_id)
   logger.info(f"Serving name: {serving_name}")
   logger.info(f"Serving ID: {serving_id}")
-  model_name = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
-  logger.debug(f"Model name: {model_name}")
+  if serving_id is None:
+    logger.error("Could not find service ID, creating a new one")
+    data = nbox_ws_v1.workspace.u(workspace_id).deployments(_method = "post", deployment_name = serving_name, deployment_description = "")
+    serving_id = data["deployment_id"]
+    logger.info(f"Serving ID: {serving_id}")
+  model_name = U.get_random_name().replace("-", "_")
+  logger.info(f"Model name: {model_name}")
 
   # zip init folder
-  zip_path = zip_to_nbox_folder(init_folder, serving_id, workspace_id, model_name = model_name)
+  # if git: -> git commit + git patch apply [+ git untracked]
+
+  zip_path = zip_to_nbox_folder(init_folder, serving_id, workspace_id, model_name = model_name, type = OT.SERVING)
   return _upload_serving_zip(zip_path, workspace_id, serving_id, serving_name, model_name)
 
 
@@ -91,7 +99,9 @@ def _upload_serving_zip(zip_path, workspace_id, serving_id, serving_name, model_
   logger.debug(f"deployment_id: {deployment_id}")
 
   # upload the file to a S3 -> don't raise for status here
-  logger.debug("Uploading model to S3 ...")
+  # TODO: @yashbonde use poster to upload files, requests doesn't support multipart uploads
+  # https://stackoverflow.com/questions/15973204/using-python-requests-to-bridge-a-file-without-loading-into-memory
+  logger.info(f"Uploading model to S3 ... (fs: {file_size/1024/1024:0.3f} MB)")
   r = requests.post(url=out["url"], data=out["fields"], files={"file": (out["fields"]["key"], open(zip_path, "rb"))})
   status = r.status_code == 204
   logger.debug(f"Upload status: {status}")
@@ -102,16 +112,16 @@ def _upload_serving_zip(zip_path, workspace_id, serving_id, serving_name, model_
 
   # write out all the commands for this deployment
   logger.info("API will soon be hosted, here's how you can use it:")
-  _api = f"Operator.from_serving('{serving_id}', $NBX_TOKEN, '{workspace_id}')"
-  _cli = f"python3 -m nbox serve forward --id_or_name '{serving_id}' --workspace_id '{workspace_id}'"
-  _curl = f"curl https://api.nimblebox.ai/{serving_id}/forward"
+  # _api = f"Operator.from_serving('{serving_id}', $NBX_TOKEN, '{workspace_id}')"
+  # _cli = f"python3 -m nbox serve forward --id_or_name '{serving_id}' --workspace_id '{workspace_id}'"
+  # _curl = f"curl https://api.nimblebox.ai/{serving_id}/forward"
   _webpage = f"{secret.get('nbx_url')}/workspace/{workspace_id}/deploy/{serving_id}"
-  logger.info(f" [python] - {_api}")
-  logger.info(f"    [CLI] - {_cli} --token $NBX_TOKEN --args")
-  logger.info(f"   [curl] - {_curl} -H 'NBX-KEY: $NBX_TOKEN' -H 'w' -d " + "'{}'")
-  logger.info(f"   [page] - {_webpage}")
+  # logger.info(f" [python] - {_api}")
+  # logger.info(f"    [CLI] - {_cli} --token $NBX_TOKEN --args")
+  # logger.info(f"   [curl] - {_curl} -H 'NBX-KEY: $NBX_TOKEN' -H 'Content-Type: application/json' -d " + "'{}'")
+  logger.info(f"  [page] - {_webpage}")
 
-  return Serve(serving_id, workspace_id)
+  return Serve(id = serving_id, model_id = model_id, workspace_id = workspace_id)
 
 
 #######################################################################################################################
@@ -149,7 +159,7 @@ def deploy_job(
   if not os.path.exists(init_folder) or not os.path.isdir(init_folder):
     raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")
 
-  job_id, job_name = _get_job_data(job_id_or_name, workspace_id)
+  job_id, job_name = _get_job_data(job_id_or_name, workspace_id = workspace_id)
   logger.info(f"Job name: {job_name}")
   logger.info(f"Job ID: {job_id}")
 
@@ -163,26 +173,27 @@ def deploy_job(
   job_proto = JobProto(
     id = job_id,
     name = job_name or U.get_random_name(True).split("-")[0],
-    created_at = get_current_timestamp(),
+    created_at = SimplerTimes.get_now_pb(),
     auth_info = NBXAuthInfo(
       username = secret.get("username"),
       workspace_id = workspace_id,
     ),
     schedule = schedule.get_message() if schedule is not None else None,
     dag = dag,
-    resource = Resource(
+    resource = resource or Resource(
       cpu = "100m",         # 100mCPU
       memory = "200Mi",     # MiB
       disk_size = "1Gi",    # GiB
-    ) if resource == None else resource,
+      max_retries = 2,      # 2 retries
+    )
   )
-  write_string_to_file(job_proto, U.join(init_folder, "job_proto.pbtxt"))
+  write_binary_to_file(job_proto, U.join(init_folder, "job_proto.msg"))
 
   if _unittest:
     return job_proto
 
   # zip the entire init folder to zip
-  zip_path = zip_to_nbox_folder(init_folder, job_id, workspace_id)
+  zip_path = zip_to_nbox_folder(init_folder, job_id, workspace_id, type = OT.JOB)
   return _upload_job_zip(zip_path, job_proto)
 
 def _upload_job_zip(zip_path: str, job_proto: JobProto):
@@ -199,7 +210,7 @@ def _upload_job_zip(zip_path: str, job_proto: JobProto):
   if not new_job:
     # incase an old job exists, we need to update few things with the new information
     logger.debug("Found existing job, checking for update masks")
-    old_job_proto = Job(job_proto.id, job_proto.auth_info.workspace_id).job_proto
+    old_job_proto = Job(job_proto.id, workspace_id = job_proto.auth_info.workspace_id).job_proto
     paths = []
     if old_job_proto.resource.SerializeToString(deterministic = True) != job_proto.resource.SerializeToString(deterministic = True):
       paths.append("resource")
@@ -225,7 +236,7 @@ def _upload_job_zip(zip_path: str, job_proto: JobProto):
   job_proto.MergeFrom(response)
   s3_url = job_proto.code.s3_url
   s3_meta = job_proto.code.s3_meta
-  logger.debug("Uploading model to S3 ...")
+  logger.info(f"Uploading model to S3 ... (fs: {job_proto.code.size/1024/1024:0.3f} MB)")
   r = requests.post(url=s3_url, data=s3_meta, files={"file": (s3_meta["key"], open(zip_path, "rb"))})
   try:
     r.raise_for_status()
@@ -235,21 +246,25 @@ def _upload_job_zip(zip_path: str, job_proto: JobProto):
 
   # if this is the first time this is being created
   if new_job:
+    job_proto.feature_gates.update({
+      "EnablePipCaching": "", # some string does not honour value
+    })
     rpc(nbox_grpc_stub.CreateJob, CreateJobRequest(job = job_proto), f"Failed to create job")
 
   # write out all the commands for this job
   logger.info("Run is now created, to 'trigger' programatically, use the following commands:")
-  _api = f"nbox.Job(id = '{job_proto.id}', workspace_id='{job_proto.auth_info.workspace_id}').trigger()"
-  _cli = f"python3 -m nbox jobs --id {job_proto.id} --workspace_id {job_proto.auth_info.workspace_id} trigger"
-  _curl = f"curl -X POST {secret.get('nbx_url')}/api/v1/workspace/{job_proto.auth_info.workspace_id}/job/{job_proto.id}/trigger"
+  # _api = f"nbox.Job(id = '{job_proto.id}', workspace_id='{job_proto.auth_info.workspace_id}').trigger()"
+  # _cli = f"python3 -m nbox jobs --id {job_proto.id} --workspace_id {job_proto.auth_info.workspace_id} trigger"
+  # _curl = f"curl -X POST {secret.get('nbx_url')}/api/v1/workspace/{job_proto.auth_info.workspace_id}/job/{job_proto.id}/trigger"
   _webpage = f"{secret.get('nbx_url')}/workspace/{job_proto.auth_info.workspace_id}/jobs/{job_proto.id}"
-  logger.info(f" [python] - {_api}")
-  logger.info(f"    [CLI] - {_cli}")
-  logger.info(f"   [curl] - {_curl} -H 'authorization: Bearer $NBX_TOKEN' -H 'Content-Type: application/json' -d " + "'{}'")
-  logger.info(f"   [page] - {_webpage}")
+  # logger.info(f" [python] - {_api}")
+  # logger.info(f"    [CLI] - {_cli}")
+  # logger.info(f"   [curl] - {_curl} -H 'authorization: Bearer $NBX_TOKEN' -H 'Content-Type: application/json' -d " + "'{}'")
+  # logger.info(f"   [page] - {_webpage}")
+  logger.info(f"See job on page: {_webpage}")
 
   # create a Job object and return so CLI can do interesting things
-  return Job(job_proto.id, job_proto.auth_info.workspace_id)
+  return Job(job_proto.id, workspace_id = job_proto.auth_info.workspace_id)
 
 
 #######################################################################################################################
@@ -260,9 +275,9 @@ Function related to both NBX-Serving and NBX-Jobs
 """
 #######################################################################################################################
 
-def zip_to_nbox_folder(init_folder, id, workspace_id, **jinja_kwargs):
+def zip_to_nbox_folder(init_folder, id, workspace_id, type, **jinja_kwargs):
   # zip all the files folder
-  all_f = U.get_files_in_folder(init_folder)
+  all_f = U.get_files_in_folder(init_folder, followlinks = True)
 
   # find a .nboxignore file and ignore items in it
   to_ignore_pat = []
@@ -278,10 +293,18 @@ def zip_to_nbox_folder(init_folder, id, workspace_id, **jinja_kwargs):
             to_ignore_pat.append(pat)
       break
 
+  # print(all_f)
+
+  # print("to_ignore_pat:", to_ignore_pat)
+  # print("to_ignore_folder:", to_ignore_folder)
+
   # two different lists for convinience
   to_remove = []
   for ignore in to_ignore_pat:
-    x = fnmatch.filter(all_f, ignore)
+    if "*" in ignore:
+      x = fnmatch.filter(all_f, ignore)
+    else:
+      x = [f for f in all_f if f.endswith(ignore)]
     to_remove.extend(x)
   to_remove_folder = []
   for ignore in to_ignore_folder:
@@ -289,9 +312,12 @@ def zip_to_nbox_folder(init_folder, id, workspace_id, **jinja_kwargs):
       if re.search(ignore, f):
         to_remove_folder.append(f)
   to_remove += to_remove_folder
-
   all_f = [x for x in all_f if x not in to_remove]
   logger.info(f"Will zip {len(all_f)} files")
+  # print(all_f)
+  # exit()
+  # print(to_remove)
+  # exit()
 
   # zip all the files folder
   zip_path = U.join(gettempdir(), f"nbxjd_{id}@{workspace_id}.nbox")
@@ -303,24 +329,30 @@ def zip_to_nbox_folder(init_folder, id, workspace_id, **jinja_kwargs):
       logger.debug(f"Zipping {f} => {arcname}")
       zip_file.write(f, arcname = arcname)
 
-    if not "exe.py" in zip_file.namelist():
-      logger.debug("exe.py already in zip")
-
+    # create the exe.py file
+    exe_jinja_path = U.join(U.folder(__file__), "assets", "exe.jinja")
+    exe_path = U.join(gettempdir(), "exe.py")
+    logger.debug(f"Writing exe to: {exe_path}")
+    with open(exe_jinja_path, "r") as f, open(exe_path, "w") as f2:
       # get a timestamp like this: Monday W34 [UTC 12 April, 2022 - 12:00:00]
       _ct = datetime.now(timezone.utc)
       _day = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][_ct.weekday()]
       created_time = f"{_day} W{_ct.isocalendar()[1]} [ UTC {_ct.strftime('%d %b, %Y - %H:%M:%S')} ]"
 
-      # create the exe.py file
-      exe_jinja_path = U.join(U.folder(__file__), "assets", "exe.jinja")
-      exe_path = U.join(gettempdir(), "exe.py")
-      logger.debug(f"Writing exe to: {exe_path}")
-      with open(exe_jinja_path, "r") as f, open(exe_path, "w") as f2:
-        f2.write(jinja2.Template(f.read()).render({
-          "created_time": created_time,
-          "nbox_version": __version__,
-          **jinja_kwargs
-        }))
-      zip_file.write(exe_path, arcname = "exe.py")
+      # fill up the jinja template
+      code = jinja2.Template(f.read()).render({
+        "created_time": created_time,
+        "nbox_version": __version__,
+        **jinja_kwargs
+      })
+      f2.write(code)
+    # print(os.stat(exe_path))
 
+    # currently the serving pod does not come with secrets file so we need to make a temporary fix while that
+    # feature is being worked on
+    if type == OT.SERVING:
+      secrets_path = U.join(U.env.NBOX_HOME_DIR(), "secrets.json")
+      zip_file.write(secrets_path, arcname = ".nbx/secrets.json")
+
+    zip_file.write(exe_path, arcname = "exe.py")
   return zip_path
